let deriver = "json"
let sanitize expr =
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_open
         (Asttypes.Override,
           {
             Asttypes.txt = (Longident.Lident "Ppx_deriving_runtime");
             Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
           }, expr));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes =
      [({
          Asttypes.txt = "ocaml.warning";
          Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
        },
         (Parsetree.PStr
            [{
               Parsetree.pstr_desc =
                 (Parsetree.Pstr_eval
                    ({
                       Parsetree.pexp_desc =
                         (Parsetree.Pexp_constant
                            (Parsetree.Pconst_string ("-A", None)));
                       Parsetree.pexp_loc =
                         (Pervasives.(!) Ast_helper.default_loc);
                       Parsetree.pexp_attributes = []
                     }, []));
               Parsetree.pstr_loc = (Pervasives.(!) Ast_helper.default_loc)
             }]))]
  }
let runtimename = "Deriving_Json"
let rt name = Ast_convenience.evar (Printf.sprintf "%s.%s" runtimename name)
let rt_t arg =
  Ast_convenience.tconstr (Printf.sprintf "%s.t" runtimename) [arg]
let lexer_ident name = Printf.sprintf "%s_lexer.%s" runtimename name
let lexbuf_t = Ast_convenience.tconstr (lexer_ident "lexbuf") []
let lexer name = Ast_convenience.evar (lexer_ident name)
let var_ptuple l = (List.map Ast_convenience.pvar l) |> Ast_helper.Pat.tuple
let map_loc f { Location.txt = txt; loc } = { Location.txt = (f txt); loc }
let suffix_lid { Location.txt = txt; loc } ~suffix  =
  let txt = Ppx_deriving.mangle_lid (`Suffix suffix) txt in
  Ast_helper.Exp.ident { txt; loc } ~loc
let suffix_decl ({ Parsetree.ptype_loc = loc;_} as d) ~suffix  =
  (let s =
     (Ppx_deriving.mangle_type_decl (`Suffix suffix) d) |> Longident.parse in
   Location.mkloc s loc) |> (Ast_helper.Exp.ident ~loc)
let suffix_decl_p ({ Parsetree.ptype_loc = loc;_} as d) ~suffix  =
  (let s = Ppx_deriving.mangle_type_decl (`Suffix suffix) d in
   Location.mkloc s loc) |> (Ast_helper.Pat.var ~loc)
let rec fresh_vars ?(acc= [])  n =
  if n <= 0
  then List.rev acc
  else
    (let acc = (Ppx_deriving.fresh_var acc) :: acc in fresh_vars ~acc (n - 1))
let label_of_constructor = map_loc (fun c -> Longident.Lident c)
let wrap_write r ~pattern  =
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_fun
         (Asttypes.Nolabel, None,
           {
             Parsetree.ppat_desc =
               (Parsetree.Ppat_var
                  {
                    Asttypes.txt = "buf";
                    Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                  });
             Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.ppat_attributes = []
           },
           {
             Parsetree.pexp_desc =
               (Parsetree.Pexp_fun (Asttypes.Nolabel, None, pattern, r));
             Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.pexp_attributes = []
           }));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
let buf_expand r =
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_fun
         (Asttypes.Nolabel, None,
           {
             Parsetree.ppat_desc =
               (Parsetree.Ppat_var
                  {
                    Asttypes.txt = "buf";
                    Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                  });
             Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.ppat_attributes = []
           }, r));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
let seqlist =
  function
  | h::l ->
      let f acc e =
        {
          Parsetree.pexp_desc = (Parsetree.Pexp_sequence (acc, e));
          Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.pexp_attributes = []
        } in
      List.fold_left f h l
  | [] ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_construct
             ({
                Asttypes.txt = (Longident.Lident "()");
                Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
              }, None));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
let check_record_fields =
  List.iter @@
    (function
     | { Parsetree.pld_mutable = Mutable;_} ->
         Location.raise_errorf "%s cannot be derived for mutable records"
           deriver
     | { pld_type = { ptyp_desc = Ptyp_poly _;_};_} ->
         Location.raise_errorf "%s cannot be derived for polymorphic records"
           deriver
     | _ -> ())
let maybe_tuple_type = function | y::[] -> y | l -> Ast_helper.Typ.tuple l
let pattern_of_record l =
  let l =
    let f { Parsetree.pld_name = pld_name;_} =
      ((label_of_constructor pld_name), (Ast_helper.Pat.var pld_name)) in
    List.map f l in
  Ast_helper.Pat.record l Asttypes.Closed
let rec write_tuple_contents l ly ~tag  ~poly  =
  let e =
    let f arg y =
      let e = write_body_of_type y ~arg ~poly in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_sequence
             ({
                Parsetree.pexp_desc =
                  (Parsetree.Pexp_apply
                     ({
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_ident
                             {
                               Asttypes.txt =
                                 (Longident.Ldot
                                    ((Longident.Lident "Buffer"),
                                      "add_string"));
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             });
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      },
                       [(Asttypes.Nolabel,
                          {
                            Parsetree.pexp_desc =
                              (Parsetree.Pexp_ident
                                 {
                                   Asttypes.txt = (Longident.Lident "buf");
                                   Asttypes.loc =
                                     (Pervasives.(!) Ast_helper.default_loc)
                                 });
                            Parsetree.pexp_loc =
                              (Pervasives.(!) Ast_helper.default_loc);
                            Parsetree.pexp_attributes = []
                          });
                       (Asttypes.Nolabel,
                         {
                           Parsetree.pexp_desc =
                             (Parsetree.Pexp_constant
                                (Parsetree.Pconst_string (",", None)));
                           Parsetree.pexp_loc =
                             (Pervasives.(!) Ast_helper.default_loc);
                           Parsetree.pexp_attributes = []
                         })]));
                Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
                Parsetree.pexp_attributes = []
              }, e));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      } in
    (List.map2 f l ly) |> seqlist
  and s = Ast_convenience.str ("[" ^ (string_of_int tag)) in
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_sequence
         ({
            Parsetree.pexp_desc =
              (Parsetree.Pexp_apply
                 ({
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt =
                             (Longident.Ldot
                                ((Longident.Lident "Buffer"), "add_string"));
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  },
                   [(Asttypes.Nolabel,
                      {
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_ident
                             {
                               Asttypes.txt = (Longident.Lident "buf");
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             });
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      });
                   (Asttypes.Nolabel, s)]));
            Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
            Parsetree.pexp_attributes = []
          },
           {
             Parsetree.pexp_desc =
               (Parsetree.Pexp_sequence
                  (e,
                    {
                      Parsetree.pexp_desc =
                        (Parsetree.Pexp_apply
                           ({
                              Parsetree.pexp_desc =
                                (Parsetree.Pexp_ident
                                   {
                                     Asttypes.txt =
                                       (Longident.Ldot
                                          ((Longident.Lident "Buffer"),
                                            "add_string"));
                                     Asttypes.loc =
                                       (Pervasives.(!) Ast_helper.default_loc)
                                   });
                              Parsetree.pexp_loc =
                                (Pervasives.(!) Ast_helper.default_loc);
                              Parsetree.pexp_attributes = []
                            },
                             [(Asttypes.Nolabel,
                                {
                                  Parsetree.pexp_desc =
                                    (Parsetree.Pexp_ident
                                       {
                                         Asttypes.txt =
                                           (Longident.Lident "buf");
                                         Asttypes.loc =
                                           (Pervasives.(!)
                                              Ast_helper.default_loc)
                                       });
                                  Parsetree.pexp_loc =
                                    (Pervasives.(!) Ast_helper.default_loc);
                                  Parsetree.pexp_attributes = []
                                });
                             (Asttypes.Nolabel,
                               {
                                 Parsetree.pexp_desc =
                                   (Parsetree.Pexp_constant
                                      (Parsetree.Pconst_string ("]", None)));
                                 Parsetree.pexp_loc =
                                   (Pervasives.(!) Ast_helper.default_loc);
                                 Parsetree.pexp_attributes = []
                               })]));
                      Parsetree.pexp_loc =
                        (Pervasives.(!) Ast_helper.default_loc);
                      Parsetree.pexp_attributes = []
                    }));
             Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.pexp_attributes = []
           }));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
and write_body_of_tuple_type l ~arg  ~poly  ~tag  =
  let n = List.length l in
  let vars = fresh_vars n in
  let e = write_tuple_contents vars l ~tag ~poly
  and p = var_ptuple vars in
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_let
         (Asttypes.Nonrecursive,
           [{
              Parsetree.pvb_pat = p;
              Parsetree.pvb_expr = arg;
              Parsetree.pvb_attributes = [];
              Parsetree.pvb_loc = (Pervasives.(!) Ast_helper.default_loc)
            }], e));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
and write_poly_case r ~arg  ~poly  =
  match r with
  | Parsetree.Rtag ({ txt = label;_}, _, _, l) ->
      let i = Ppx_deriving.hash_variant label
      and n = List.length l in
      let v = Ppx_deriving.fresh_var [] in
      let lhs =
        (if n = 0 then None else Some (Ast_convenience.pvar v)) |>
          (Ast_helper.Pat.variant label)
      and rhs =
        match l with
        | [] ->
            let e = Ast_convenience.int i in
            {
              Parsetree.pexp_desc =
                (Parsetree.Pexp_apply
                   ((rt "Json_int.write"),
                     [(Asttypes.Nolabel,
                        {
                          Parsetree.pexp_desc =
                            (Parsetree.Pexp_ident
                               {
                                 Asttypes.txt = (Longident.Lident "buf");
                                 Asttypes.loc =
                                   (Pervasives.(!) Ast_helper.default_loc)
                               });
                          Parsetree.pexp_loc =
                            (Pervasives.(!) Ast_helper.default_loc);
                          Parsetree.pexp_attributes = []
                        });
                     (Asttypes.Nolabel, e)]));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            }
        | _ ->
            let l =
              [{
                 Parsetree.ptyp_desc =
                   (Parsetree.Ptyp_constr
                      ({
                         Asttypes.txt = (Longident.Lident "int");
                         Asttypes.loc =
                           (Pervasives.(!) Ast_helper.default_loc)
                       }, []));
                 Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
                 Parsetree.ptyp_attributes = []
               };
              maybe_tuple_type l]
            and arg =
              Ast_helper.Exp.tuple
                (let open Ast_convenience in [int i; evar v]) in
            write_body_of_tuple_type l ~arg ~poly ~tag:0 in
      Ast_helper.Exp.case lhs rhs
  | Rinherit ({ ptyp_desc = Ptyp_constr (lid, _); ptyp_loc;_} as y) ->
      Ast_helper.Exp.case
        ((let open Ast_helper.Pat in alias (type_ lid))
           (Location.mkloc arg ptyp_loc)) (write_body_of_type y ~arg ~poly)
  | Rinherit { ptyp_loc;_} ->
      Location.raise_errorf ~loc:ptyp_loc "%s write case cannot be derived"
        deriver
and write_body_of_type y ~arg:(arg : string)  ~poly  =
  let arg = Ast_convenience.evar arg
  and arg' = arg in
  match y with
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "unit"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_unit.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int32"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({ Asttypes.txt = Longident.Ldot (Longident.Lident "Int32", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int32.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int64"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({ Asttypes.txt = Longident.Ldot (Longident.Lident "Int64", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int64.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "nativeint"; Asttypes.loc = _ },
         []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({
            Asttypes.txt = Longident.Ldot (Longident.Lident "Nativeint", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_nativeint.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "float"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_float.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "bool"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_bool.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "char"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_char.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "string"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_string.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "bytes"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_bytes.write"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "list"; Asttypes.loc = _ }, y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      let e = write_of_type y ~poly in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "write_list"),
               [(Asttypes.Nolabel, e);
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "ref"; Asttypes.loc = _ }, y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      let e = write_of_type y ~poly in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "write_ref"),
               [(Asttypes.Nolabel, e);
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "option"; Asttypes.loc = _ },
         y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      let e = write_of_type y ~poly in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "write_option"),
               [(Asttypes.Nolabel, e);
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "array"; Asttypes.loc = _ },
         y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      let e = write_of_type y ~poly in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "write_array"),
               [(Asttypes.Nolabel, e);
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_desc = Ptyp_var v;_} when poly ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((Ast_convenience.evar ("poly_" ^ v)),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_desc = Ptyp_tuple l;_} ->
      write_body_of_tuple_type l ~arg ~poly ~tag:0
  | { Parsetree.ptyp_desc = Ptyp_variant (l, _, _);_} ->
      Ast_helper.Exp.match_ arg
        (List.map (write_poly_case ~arg:arg' ~poly) l)
  | { Parsetree.ptyp_desc = Ptyp_constr (lid, l);_} ->
      let e = suffix_lid lid ~suffix:"to_json"
      and l = List.map (write_of_type ~poly) l in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((Ast_convenience.app e l),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, arg)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_loc = ptyp_loc;_} ->
      Location.raise_errorf ~loc:ptyp_loc "%s_write cannot be derived for %s"
        deriver (Ppx_deriving.string_of_core_type y)
and write_of_type y ~poly  =
  let arg = "a" in
  let pattern = Ast_convenience.pvar arg in
  wrap_write (write_body_of_type y ~arg ~poly) ~pattern
and write_body_of_record ~tag  l =
  let l = let f { Parsetree.pld_name = { txt;_};_} = txt in List.map f l
  and ly =
    let f { Parsetree.pld_type = pld_type;_} = pld_type in List.map f l in
  write_tuple_contents l ly ~tag ~poly:true
and write_of_record ?(tag= 0)  _d l =
  let pattern = pattern_of_record l
  and e = write_body_of_record ~tag l in wrap_write e ~pattern
let recognize_case_of_constructor i l =
  let lhs =
    match l with
    | [] ->
        {
          Parsetree.ppat_desc =
            (Parsetree.Ppat_variant ("Cst", (Some (Ast_convenience.pint i))));
          Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.ppat_attributes = []
        }
    | _ ->
        {
          Parsetree.ppat_desc =
            (Parsetree.Ppat_variant ("NCst", (Some (Ast_convenience.pint i))));
          Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.ppat_attributes = []
        } in
  Ast_helper.Exp.case lhs
    {
      Parsetree.pexp_desc =
        (Parsetree.Pexp_construct
           ({
              Asttypes.txt = (Longident.Lident "true");
              Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
            }, None));
      Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.pexp_attributes = []
    }
let recognize_body_of_poly_variant l ~loc  =
  let l =
    let f =
      function
      | Parsetree.Rtag ({ txt = label;_}, _, _, l) ->
          let i = Ppx_deriving.hash_variant label in
          recognize_case_of_constructor i l
      | Rinherit { ptyp_desc = Ptyp_constr (lid, _);_} ->
          let guard =
            {
              Parsetree.pexp_desc =
                (Parsetree.Pexp_apply
                   ((suffix_lid lid ~suffix:"recognize"),
                     [(Asttypes.Nolabel,
                        {
                          Parsetree.pexp_desc =
                            (Parsetree.Pexp_ident
                               {
                                 Asttypes.txt = (Longident.Lident "x");
                                 Asttypes.loc =
                                   (Pervasives.(!) Ast_helper.default_loc)
                               });
                          Parsetree.pexp_loc =
                            (Pervasives.(!) Ast_helper.default_loc);
                          Parsetree.pexp_attributes = []
                        })]));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            } in
          Ast_helper.Exp.case ~guard
            {
              Parsetree.ppat_desc =
                (Parsetree.Ppat_var
                   {
                     Asttypes.txt = "x";
                     Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                   });
              Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.ppat_attributes = []
            }
            {
              Parsetree.pexp_desc =
                (Parsetree.Pexp_construct
                   ({
                      Asttypes.txt = (Longident.Lident "true");
                      Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                    }, None));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            }
      | _ ->
          Location.raise_errorf ~loc "%s_recognize cannot be derived" deriver
    and default =
      Ast_helper.Exp.case
        {
          Parsetree.ppat_desc = Parsetree.Ppat_any;
          Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.ppat_attributes = []
        }
        {
          Parsetree.pexp_desc =
            (Parsetree.Pexp_construct
               ({
                  Asttypes.txt = (Longident.Lident "false");
                  Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                }, None));
          Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.pexp_attributes = []
        } in
    (List.map f l) @ [default] in
  Ast_helper.Exp.function_ l
let tag_error_case ?(typename= "")  () =
  let y = Ast_convenience.str typename in
  Ast_helper.Exp.case
    {
      Parsetree.ppat_desc = Parsetree.Ppat_any;
      Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ppat_attributes = []
    }
    {
      Parsetree.pexp_desc =
        (Parsetree.Pexp_apply
           ((lexer "tag_error"),
             [((Asttypes.Labelled "typename"), y);
             (Asttypes.Nolabel,
               {
                 Parsetree.pexp_desc =
                   (Parsetree.Pexp_ident
                      {
                        Asttypes.txt = (Longident.Lident "buf");
                        Asttypes.loc =
                          (Pervasives.(!) Ast_helper.default_loc)
                      });
                 Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
                 Parsetree.pexp_attributes = []
               })]));
      Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.pexp_attributes = []
    }
let maybe_tuple_type = function | y::[] -> y | l -> Ast_helper.Typ.tuple l
let rec read_poly_case ?decl  y =
  function
  | Parsetree.Rtag ({ txt = label;_}, _, _, l) ->
      let i = (Ppx_deriving.hash_variant label) |> Ast_convenience.pint in
      (match l with
       | [] ->
           Ast_helper.Exp.case
             {
               Parsetree.ppat_desc =
                 (Parsetree.Ppat_variant ("Cst", (Some i)));
               Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ppat_attributes = []
             } (Ast_helper.Exp.variant label None)
       | l ->
           Ast_helper.Exp.case
             {
               Parsetree.ppat_desc =
                 (Parsetree.Ppat_variant ("NCst", (Some i)));
               Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ppat_attributes = []
             }
             {
               Parsetree.pexp_desc =
                 (Parsetree.Pexp_sequence
                    ({
                       Parsetree.pexp_desc =
                         (Parsetree.Pexp_apply
                            ((lexer "read_comma"),
                              [(Asttypes.Nolabel,
                                 {
                                   Parsetree.pexp_desc =
                                     (Parsetree.Pexp_ident
                                        {
                                          Asttypes.txt =
                                            (Longident.Lident "buf");
                                          Asttypes.loc =
                                            (Pervasives.(!)
                                               Ast_helper.default_loc)
                                        });
                                   Parsetree.pexp_loc =
                                     (Pervasives.(!) Ast_helper.default_loc);
                                   Parsetree.pexp_attributes = []
                                 })]));
                       Parsetree.pexp_loc =
                         (Pervasives.(!) Ast_helper.default_loc);
                       Parsetree.pexp_attributes = []
                     },
                      {
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_let
                             (Asttypes.Nonrecursive,
                               [{
                                  Parsetree.pvb_pat =
                                    {
                                      Parsetree.ppat_desc =
                                        (Parsetree.Ppat_var
                                           {
                                             Asttypes.txt = "v";
                                             Asttypes.loc =
                                               (Pervasives.(!)
                                                  Ast_helper.default_loc)
                                           });
                                      Parsetree.ppat_loc =
                                        (Pervasives.(!)
                                           Ast_helper.default_loc);
                                      Parsetree.ppat_attributes = []
                                    };
                                  Parsetree.pvb_expr =
                                    (read_body_of_type ?decl
                                       (maybe_tuple_type l));
                                  Parsetree.pvb_attributes = [];
                                  Parsetree.pvb_loc =
                                    (Pervasives.(!) Ast_helper.default_loc)
                                }],
                               {
                                 Parsetree.pexp_desc =
                                   (Parsetree.Pexp_sequence
                                      ({
                                         Parsetree.pexp_desc =
                                           (Parsetree.Pexp_apply
                                              ((lexer "read_rbracket"),
                                                [(Asttypes.Nolabel,
                                                   {
                                                     Parsetree.pexp_desc =
                                                       (Parsetree.Pexp_ident
                                                          {
                                                            Asttypes.txt =
                                                              (Longident.Lident
                                                                 "buf");
                                                            Asttypes.loc =
                                                              (Pervasives.(!)
                                                                 Ast_helper.default_loc)
                                                          });
                                                     Parsetree.pexp_loc =
                                                       (Pervasives.(!)
                                                          Ast_helper.default_loc);
                                                     Parsetree.pexp_attributes
                                                       = []
                                                   })]));
                                         Parsetree.pexp_loc =
                                           (Pervasives.(!)
                                              Ast_helper.default_loc);
                                         Parsetree.pexp_attributes = []
                                       },
                                        (Ast_helper.Exp.variant label
                                           (Some
                                              {
                                                Parsetree.pexp_desc =
                                                  (Parsetree.Pexp_ident
                                                     {
                                                       Asttypes.txt =
                                                         (Longident.Lident
                                                            "v");
                                                       Asttypes.loc =
                                                         (Pervasives.(!)
                                                            Ast_helper.default_loc)
                                                     });
                                                Parsetree.pexp_loc =
                                                  (Pervasives.(!)
                                                     Ast_helper.default_loc);
                                                Parsetree.pexp_attributes =
                                                  []
                                              }))));
                                 Parsetree.pexp_loc =
                                   (Pervasives.(!) Ast_helper.default_loc);
                                 Parsetree.pexp_attributes = []
                               }));
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      }));
               Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.pexp_attributes = []
             })
  | Rinherit { ptyp_desc = Ptyp_constr (lid, l);_} ->
      let guard =
        {
          Parsetree.pexp_desc =
            (Parsetree.Pexp_apply
               ((suffix_lid lid ~suffix:"recognize"),
                 [(Asttypes.Nolabel,
                    {
                      Parsetree.pexp_desc =
                        (Parsetree.Pexp_ident
                           {
                             Asttypes.txt = (Longident.Lident "x");
                             Asttypes.loc =
                               (Pervasives.(!) Ast_helper.default_loc)
                           });
                      Parsetree.pexp_loc =
                        (Pervasives.(!) Ast_helper.default_loc);
                      Parsetree.pexp_attributes = []
                    })]));
          Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.pexp_attributes = []
        }
      and e =
        let e = suffix_lid lid ~suffix:"of_json_with_tag"
        and l = List.map (read_of_type ?decl) l in
        {
          Parsetree.pexp_desc =
            (Parsetree.Pexp_coerce
               ({
                  Parsetree.pexp_desc =
                    (Parsetree.Pexp_apply
                       ((Ast_convenience.app e l),
                         [(Asttypes.Nolabel,
                            {
                              Parsetree.pexp_desc =
                                (Parsetree.Pexp_ident
                                   {
                                     Asttypes.txt = (Longident.Lident "buf");
                                     Asttypes.loc =
                                       (Pervasives.(!) Ast_helper.default_loc)
                                   });
                              Parsetree.pexp_loc =
                                (Pervasives.(!) Ast_helper.default_loc);
                              Parsetree.pexp_attributes = []
                            });
                         (Asttypes.Nolabel,
                           {
                             Parsetree.pexp_desc =
                               (Parsetree.Pexp_ident
                                  {
                                    Asttypes.txt = (Longident.Lident "x");
                                    Asttypes.loc =
                                      (Pervasives.(!) Ast_helper.default_loc)
                                  });
                             Parsetree.pexp_loc =
                               (Pervasives.(!) Ast_helper.default_loc);
                             Parsetree.pexp_attributes = []
                           })]));
                  Parsetree.pexp_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.pexp_attributes = []
                }, None, y));
          Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.pexp_attributes = []
        } in
      Ast_helper.Exp.case ~guard
        {
          Parsetree.ppat_desc =
            (Parsetree.Ppat_var
               {
                 Asttypes.txt = "x";
                 Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
               });
          Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.ppat_attributes = []
        } e
  | Rinherit { ptyp_loc;_} ->
      Location.raise_errorf ~loc:ptyp_loc "%s read case cannot be derived"
        deriver
and read_of_poly_variant ?decl  l y ~loc:_  =
  (((List.map (read_poly_case ?decl y) l) @ [tag_error_case ()]) |>
     Ast_helper.Exp.function_)
    |> buf_expand
and read_tuple_contents ?decl  l ~f  =
  let n = List.length l in
  let lv = fresh_vars n in
  let f v y acc =
    let e = read_body_of_type ?decl y in
    {
      Parsetree.pexp_desc =
        (Parsetree.Pexp_sequence
           ({
              Parsetree.pexp_desc =
                (Parsetree.Pexp_apply
                   ((lexer "read_comma"),
                     [(Asttypes.Nolabel,
                        {
                          Parsetree.pexp_desc =
                            (Parsetree.Pexp_ident
                               {
                                 Asttypes.txt = (Longident.Lident "buf");
                                 Asttypes.loc =
                                   (Pervasives.(!) Ast_helper.default_loc)
                               });
                          Parsetree.pexp_loc =
                            (Pervasives.(!) Ast_helper.default_loc);
                          Parsetree.pexp_attributes = []
                        })]));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            },
             {
               Parsetree.pexp_desc =
                 (Parsetree.Pexp_let
                    (Asttypes.Nonrecursive,
                      [{
                         Parsetree.pvb_pat = (Ast_convenience.pvar v);
                         Parsetree.pvb_expr = e;
                         Parsetree.pvb_attributes = [];
                         Parsetree.pvb_loc =
                           (Pervasives.(!) Ast_helper.default_loc)
                       }], acc));
               Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.pexp_attributes = []
             }));
      Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.pexp_attributes = []
    }
  and acc = (List.map Ast_convenience.evar lv) |> f in
  let acc =
    {
      Parsetree.pexp_desc =
        (Parsetree.Pexp_sequence
           ({
              Parsetree.pexp_desc =
                (Parsetree.Pexp_apply
                   ((lexer "read_rbracket"),
                     [(Asttypes.Nolabel,
                        {
                          Parsetree.pexp_desc =
                            (Parsetree.Pexp_ident
                               {
                                 Asttypes.txt = (Longident.Lident "buf");
                                 Asttypes.loc =
                                   (Pervasives.(!) Ast_helper.default_loc)
                               });
                          Parsetree.pexp_loc =
                            (Pervasives.(!) Ast_helper.default_loc);
                          Parsetree.pexp_attributes = []
                        })]));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            }, acc));
      Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.pexp_attributes = []
    } in
  List.fold_right2 f lv l acc
and read_body_of_tuple_type ?decl  l =
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_sequence
         ({
            Parsetree.pexp_desc =
              (Parsetree.Pexp_apply
                 ((lexer "read_lbracket"),
                   [(Asttypes.Nolabel,
                      {
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_ident
                             {
                               Asttypes.txt = (Longident.Lident "buf");
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             });
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      })]));
            Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
            Parsetree.pexp_attributes = []
          },
           {
             Parsetree.pexp_desc =
               (Parsetree.Pexp_sequence
                  ({
                     Parsetree.pexp_desc =
                       (Parsetree.Pexp_apply
                          ({
                             Parsetree.pexp_desc =
                               (Parsetree.Pexp_ident
                                  {
                                    Asttypes.txt =
                                      (Longident.Lident "ignore");
                                    Asttypes.loc =
                                      (Pervasives.(!) Ast_helper.default_loc)
                                  });
                             Parsetree.pexp_loc =
                               (Pervasives.(!) Ast_helper.default_loc);
                             Parsetree.pexp_attributes = []
                           },
                            [(Asttypes.Nolabel,
                               {
                                 Parsetree.pexp_desc =
                                   (Parsetree.Pexp_apply
                                      ((lexer "read_tag_1"),
                                        [(Asttypes.Nolabel,
                                           {
                                             Parsetree.pexp_desc =
                                               (Parsetree.Pexp_constant
                                                  (Parsetree.Pconst_integer
                                                     ("0", None)));
                                             Parsetree.pexp_loc =
                                               (Pervasives.(!)
                                                  Ast_helper.default_loc);
                                             Parsetree.pexp_attributes = []
                                           });
                                        (Asttypes.Nolabel,
                                          {
                                            Parsetree.pexp_desc =
                                              (Parsetree.Pexp_ident
                                                 {
                                                   Asttypes.txt =
                                                     (Longident.Lident "buf");
                                                   Asttypes.loc =
                                                     (Pervasives.(!)
                                                        Ast_helper.default_loc)
                                                 });
                                            Parsetree.pexp_loc =
                                              (Pervasives.(!)
                                                 Ast_helper.default_loc);
                                            Parsetree.pexp_attributes = []
                                          })]));
                                 Parsetree.pexp_loc =
                                   (Pervasives.(!) Ast_helper.default_loc);
                                 Parsetree.pexp_attributes = []
                               })]));
                     Parsetree.pexp_loc =
                       (Pervasives.(!) Ast_helper.default_loc);
                     Parsetree.pexp_attributes = []
                   }, (read_tuple_contents ?decl l ~f:Ast_helper.Exp.tuple)));
             Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.pexp_attributes = []
           }));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
and read_of_record_raw ?decl  ?(return= fun x -> x)  l =
  let f =
    let f { Parsetree.pld_name = pld_name;_} e =
      ((label_of_constructor pld_name), e) in
    fun l' -> return (Ast_helper.Exp.record (List.map2 f l l') None)
  and l = let f { Parsetree.pld_type = pld_type;_} = pld_type in List.map f l in
  read_tuple_contents l ?decl ~f
and read_of_record decl l =
  let e = read_of_record_raw ~decl l in
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_sequence
         ({
            Parsetree.pexp_desc =
              (Parsetree.Pexp_apply
                 ((lexer "read_lbracket"),
                   [(Asttypes.Nolabel,
                      {
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_ident
                             {
                               Asttypes.txt = (Longident.Lident "buf");
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             });
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      })]));
            Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
            Parsetree.pexp_attributes = []
          },
           {
             Parsetree.pexp_desc =
               (Parsetree.Pexp_sequence
                  ({
                     Parsetree.pexp_desc =
                       (Parsetree.Pexp_apply
                          ({
                             Parsetree.pexp_desc =
                               (Parsetree.Pexp_ident
                                  {
                                    Asttypes.txt =
                                      (Longident.Lident "ignore");
                                    Asttypes.loc =
                                      (Pervasives.(!) Ast_helper.default_loc)
                                  });
                             Parsetree.pexp_loc =
                               (Pervasives.(!) Ast_helper.default_loc);
                             Parsetree.pexp_attributes = []
                           },
                            [(Asttypes.Nolabel,
                               {
                                 Parsetree.pexp_desc =
                                   (Parsetree.Pexp_apply
                                      ((lexer "read_tag_2"),
                                        [(Asttypes.Nolabel,
                                           {
                                             Parsetree.pexp_desc =
                                               (Parsetree.Pexp_constant
                                                  (Parsetree.Pconst_integer
                                                     ("0", None)));
                                             Parsetree.pexp_loc =
                                               (Pervasives.(!)
                                                  Ast_helper.default_loc);
                                             Parsetree.pexp_attributes = []
                                           });
                                        (Asttypes.Nolabel,
                                          {
                                            Parsetree.pexp_desc =
                                              (Parsetree.Pexp_constant
                                                 (Parsetree.Pconst_integer
                                                    ("254", None)));
                                            Parsetree.pexp_loc =
                                              (Pervasives.(!)
                                                 Ast_helper.default_loc);
                                            Parsetree.pexp_attributes = []
                                          });
                                        (Asttypes.Nolabel,
                                          {
                                            Parsetree.pexp_desc =
                                              (Parsetree.Pexp_ident
                                                 {
                                                   Asttypes.txt =
                                                     (Longident.Lident "buf");
                                                   Asttypes.loc =
                                                     (Pervasives.(!)
                                                        Ast_helper.default_loc)
                                                 });
                                            Parsetree.pexp_loc =
                                              (Pervasives.(!)
                                                 Ast_helper.default_loc);
                                            Parsetree.pexp_attributes = []
                                          })]));
                                 Parsetree.pexp_loc =
                                   (Pervasives.(!) Ast_helper.default_loc);
                                 Parsetree.pexp_attributes = []
                               })]));
                     Parsetree.pexp_loc =
                       (Pervasives.(!) Ast_helper.default_loc);
                     Parsetree.pexp_attributes = []
                   }, e));
             Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
             Parsetree.pexp_attributes = []
           }));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  } |> buf_expand
and read_body_of_type ?decl  y =
  let poly = match decl with | Some _ -> true | _ -> false in
  match y with
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "unit"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_unit.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int32"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({ Asttypes.txt = Longident.Ldot (Longident.Lident "Int32", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int32.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "int64"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({ Asttypes.txt = Longident.Ldot (Longident.Lident "Int64", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_int64.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "nativeint"; Asttypes.loc = _ },
         []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
    |{
       Parsetree.ptyp_desc = Parsetree.Ptyp_constr
         ({
            Asttypes.txt = Longident.Ldot (Longident.Lident "Nativeint", "t");
            Asttypes.loc = _ },
          []);
       Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ }
      ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_nativeint.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "float"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_float.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "bool"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_bool.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "char"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_char.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "string"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_string.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "bytes"; Asttypes.loc = _ }, []);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "Json_bytes.read"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "list"; Asttypes.loc = _ }, y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "read_list"),
               [(Asttypes.Nolabel, (read_of_type ?decl y));
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "ref"; Asttypes.loc = _ }, y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "read_ref"),
               [(Asttypes.Nolabel, (read_of_type ?decl y));
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "option"; Asttypes.loc = _ },
         y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "read_option"),
               [(Asttypes.Nolabel, (read_of_type ?decl y));
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | {
      Parsetree.ptyp_desc = Parsetree.Ptyp_constr
        ({ Asttypes.txt = Longident.Lident "array"; Asttypes.loc = _ },
         y::[]);
      Parsetree.ptyp_loc = _; Parsetree.ptyp_attributes = _ } ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((rt "read_array"),
               [(Asttypes.Nolabel, (read_of_type ?decl y));
               (Asttypes.Nolabel,
                 {
                   Parsetree.pexp_desc =
                     (Parsetree.Pexp_ident
                        {
                          Asttypes.txt = (Longident.Lident "buf");
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        });
                   Parsetree.pexp_loc =
                     (Pervasives.(!) Ast_helper.default_loc);
                   Parsetree.pexp_attributes = []
                 })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_desc = Ptyp_tuple l;_} ->
      read_body_of_tuple_type l ?decl
  | { Parsetree.ptyp_desc = Ptyp_variant (l, _, _); ptyp_loc = loc;_} ->
      let e =
        match decl with
        | Some decl ->
            let e = suffix_decl decl ~suffix:"of_json_with_tag"
            and l =
              let { Parsetree.ptype_params = l;_} = decl
              and f (y, _) = read_of_type y ~decl in List.map f l in
            Ast_convenience.app e l
        | None -> read_of_poly_variant l y ~loc
      and tag =
        {
          Parsetree.pexp_desc =
            (Parsetree.Pexp_apply
               ((lexer "read_vcase"),
                 [(Asttypes.Nolabel,
                    {
                      Parsetree.pexp_desc =
                        (Parsetree.Pexp_ident
                           {
                             Asttypes.txt = (Longident.Lident "buf");
                             Asttypes.loc =
                               (Pervasives.(!) Ast_helper.default_loc)
                           });
                      Parsetree.pexp_loc =
                        (Pervasives.(!) Ast_helper.default_loc);
                      Parsetree.pexp_attributes = []
                    })]));
          Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.pexp_attributes = []
        } in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             (e,
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  });
               (Asttypes.Nolabel, tag)]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_desc = Ptyp_var v;_} when poly ->
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((Ast_convenience.evar ("poly_" ^ v)),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_desc = Ptyp_constr (lid, l);_} ->
      let e = suffix_lid lid ~suffix:"of_json"
      and l = List.map (read_of_type ?decl) l in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((Ast_convenience.app e l),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
  | { Parsetree.ptyp_loc = ptyp_loc;_} ->
      Location.raise_errorf ~loc:ptyp_loc "%s_read cannot be derived for %s"
        deriver (Ppx_deriving.string_of_core_type y)
and read_of_type ?decl  y = (read_body_of_type ?decl y) |> buf_expand
let json_of_type ?decl  y =
  let read = read_of_type ?decl y
  and write =
    let poly = match decl with | Some _ -> true | _ -> false in
    write_of_type y ~poly in
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_apply
         ((rt "make"), [(Asttypes.Nolabel, write); (Asttypes.Nolabel, read)]));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  }
let fun_str_wrap d e y ~f  ~suffix  =
  let e = (Ppx_deriving.poly_fun_of_type_decl d e) |> sanitize
  and v = suffix_decl_p d ~suffix
  and y = Ppx_deriving.poly_arrow_of_type_decl f d y in
  let open Ast_helper in Vb.mk (Pat.constraint_ v y) e
let read_str_wrap d e =
  let f y =
    {
      Parsetree.ptyp_desc =
        (Parsetree.Ptyp_arrow (Asttypes.Nolabel, lexbuf_t, y));
      Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ptyp_attributes = []
    }
  and suffix = "of_json" in
  let y = f (Ppx_deriving.core_type_of_type_decl d) in
  fun_str_wrap d e y ~f ~suffix
let read_tag_str_wrap d e =
  let f y =
    {
      Parsetree.ptyp_desc =
        (Parsetree.Ptyp_arrow (Asttypes.Nolabel, lexbuf_t, y));
      Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ptyp_attributes = []
    }
  and suffix = "of_json_with_tag"
  and y =
    let y = Ppx_deriving.core_type_of_type_decl d in
    {
      Parsetree.ptyp_desc =
        (Parsetree.Ptyp_arrow
           (Asttypes.Nolabel, lexbuf_t,
             {
               Parsetree.ptyp_desc =
                 (Parsetree.Ptyp_arrow
                    (Asttypes.Nolabel,
                      {
                        Parsetree.ptyp_desc =
                          (Parsetree.Ptyp_variant
                             ([Parsetree.Rtag
                                 ({
                                    Asttypes.txt = "NCst";
                                    Asttypes.loc =
                                      (Pervasives.(!) Ast_helper.default_loc)
                                  }, [], false,
                                   [{
                                      Parsetree.ptyp_desc =
                                        (Parsetree.Ptyp_constr
                                           ({
                                              Asttypes.txt =
                                                (Longident.Lident "int");
                                              Asttypes.loc =
                                                (Pervasives.(!)
                                                   Ast_helper.default_loc)
                                            }, []));
                                      Parsetree.ptyp_loc =
                                        (Pervasives.(!)
                                           Ast_helper.default_loc);
                                      Parsetree.ptyp_attributes = []
                                    }]);
                              Parsetree.Rtag
                                ({
                                   Asttypes.txt = "Cst";
                                   Asttypes.loc =
                                     (Pervasives.(!) Ast_helper.default_loc)
                                 }, [], false,
                                  [{
                                     Parsetree.ptyp_desc =
                                       (Parsetree.Ptyp_constr
                                          ({
                                             Asttypes.txt =
                                               (Longident.Lident "int");
                                             Asttypes.loc =
                                               (Pervasives.(!)
                                                  Ast_helper.default_loc)
                                           }, []));
                                     Parsetree.ptyp_loc =
                                       (Pervasives.(!) Ast_helper.default_loc);
                                     Parsetree.ptyp_attributes = []
                                   }])], Asttypes.Closed, None));
                        Parsetree.ptyp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.ptyp_attributes = []
                      }, y));
               Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ptyp_attributes = []
             }));
      Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ptyp_attributes = []
    } in
  fun_str_wrap d e y ~f ~suffix
let write_str_wrap d e =
  let f y =
    {
      Parsetree.ptyp_desc =
        (Parsetree.Ptyp_arrow
           (Asttypes.Nolabel,
             {
               Parsetree.ptyp_desc =
                 (Parsetree.Ptyp_constr
                    ({
                       Asttypes.txt =
                         (Longident.Ldot ((Longident.Lident "Buffer"), "t"));
                       Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                     }, []));
               Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ptyp_attributes = []
             },
             {
               Parsetree.ptyp_desc =
                 (Parsetree.Ptyp_arrow
                    (Asttypes.Nolabel, y,
                      {
                        Parsetree.ptyp_desc =
                          (Parsetree.Ptyp_constr
                             ({
                                Asttypes.txt = (Longident.Lident "unit");
                                Asttypes.loc =
                                  (Pervasives.(!) Ast_helper.default_loc)
                              }, []));
                        Parsetree.ptyp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.ptyp_attributes = []
                      }));
               Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ptyp_attributes = []
             }));
      Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ptyp_attributes = []
    }
  and suffix = "to_json" in
  let y =
    let y = Ppx_deriving.core_type_of_type_decl d in
    (match d with
     | {
         ptype_manifest = Some
           { ptyp_desc = Parsetree.Ptyp_variant (_, _, _);_};_}
         ->
         {
           Parsetree.ptyp_desc =
             (Parsetree.Ptyp_variant
                ([Parsetree.Rinherit y], Asttypes.Open, None));
           Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
           Parsetree.ptyp_attributes = []
         }
     | _ -> y) |> f in
  fun_str_wrap d e y ~f ~suffix
let recognize_str_wrap d e =
  let v = suffix_decl_p d ~suffix:"recognize"
  and y =
    {
      Parsetree.ptyp_desc =
        (Parsetree.Ptyp_arrow
           (Asttypes.Nolabel,
             {
               Parsetree.ptyp_desc =
                 (Parsetree.Ptyp_variant
                    ([Parsetree.Rtag
                        ({
                           Asttypes.txt = "NCst";
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         }, [], false,
                          [{
                             Parsetree.ptyp_desc =
                               (Parsetree.Ptyp_constr
                                  ({
                                     Asttypes.txt = (Longident.Lident "int");
                                     Asttypes.loc =
                                       (Pervasives.(!) Ast_helper.default_loc)
                                   }, []));
                             Parsetree.ptyp_loc =
                               (Pervasives.(!) Ast_helper.default_loc);
                             Parsetree.ptyp_attributes = []
                           }]);
                     Parsetree.Rtag
                       ({
                          Asttypes.txt = "Cst";
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        }, [], false,
                         [{
                            Parsetree.ptyp_desc =
                              (Parsetree.Ptyp_constr
                                 ({
                                    Asttypes.txt = (Longident.Lident "int");
                                    Asttypes.loc =
                                      (Pervasives.(!) Ast_helper.default_loc)
                                  }, []));
                            Parsetree.ptyp_loc =
                              (Pervasives.(!) Ast_helper.default_loc);
                            Parsetree.ptyp_attributes = []
                          }])], Asttypes.Closed, None));
               Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ptyp_attributes = []
             },
             {
               Parsetree.ptyp_desc =
                 (Parsetree.Ptyp_constr
                    ({
                       Asttypes.txt = (Longident.Lident "bool");
                       Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                     }, []));
               Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
               Parsetree.ptyp_attributes = []
             }));
      Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
      Parsetree.ptyp_attributes = []
    } in
  let open Ast_helper in Vb.mk (Pat.constraint_ v y) e
let json_poly_type d =
  let f y = rt_t y in
  let y = f (Ppx_deriving.core_type_of_type_decl d) in
  Ppx_deriving.poly_arrow_of_type_decl f d y
let json_str_wrap d e =
  let v = suffix_decl_p d ~suffix:"json"
  and e = let open Ppx_deriving in poly_fun_of_type_decl d e
  and y = json_poly_type d in
  let open Ast_helper in Vb.mk (Pat.constraint_ v y) e
let json_str d =
  let write =
    let f acc id =
      let id = id.Location.txt in
      let poly = Ast_convenience.evar ("poly_" ^ id) in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             (acc,
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_apply
                         ((rt "write"), [(Asttypes.Nolabel, poly)]));
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
    and acc = suffix_decl d ~suffix:"to_json" in
    Ppx_deriving.fold_left_type_decl f acc d
  and read =
    let f acc id =
      let id = id.Location.txt in
      let poly = Ast_convenience.evar ("poly_" ^ id) in
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             (acc,
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_apply
                         ((rt "read"), [(Asttypes.Nolabel, poly)]));
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      }
    and acc = suffix_decl d ~suffix:"of_json" in
    Ppx_deriving.fold_left_type_decl f acc d in
  {
    Parsetree.pexp_desc =
      (Parsetree.Pexp_apply
         ((rt "make"), [(Asttypes.Nolabel, write); (Asttypes.Nolabel, read)]));
    Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
    Parsetree.pexp_attributes = []
  } |> (json_str_wrap d)
let write_decl_of_type d y =
  (let e = write_body_of_type y ~arg:"a" ~poly:true in
   {
     Parsetree.pexp_desc =
       (Parsetree.Pexp_fun
          (Asttypes.Nolabel, None,
            {
              Parsetree.ppat_desc =
                (Parsetree.Ppat_var
                   {
                     Asttypes.txt = "buf";
                     Asttypes.loc = (Pervasives.(!) Ast_helper.default_loc)
                   });
              Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.ppat_attributes = []
            },
            {
              Parsetree.pexp_desc =
                (Parsetree.Pexp_fun
                   (Asttypes.Nolabel, None,
                     {
                       Parsetree.ppat_desc =
                         (Parsetree.Ppat_var
                            {
                              Asttypes.txt = "a";
                              Asttypes.loc =
                                (Pervasives.(!) Ast_helper.default_loc)
                            });
                       Parsetree.ppat_loc =
                         (Pervasives.(!) Ast_helper.default_loc);
                       Parsetree.ppat_attributes = []
                     }, e));
              Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.pexp_attributes = []
            }));
     Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
     Parsetree.pexp_attributes = []
   }) |> (write_str_wrap d)
let read_decl_of_type decl y =
  ((read_body_of_type y ~decl) |> buf_expand) |> (read_str_wrap decl)
let json_decls_of_type decl y =
  let (recognize, read_tag) =
    match y with
    | { Parsetree.ptyp_desc = Ptyp_variant (l, _, _); ptyp_loc = loc;_} ->
        ((Some
            ((recognize_body_of_poly_variant l ~loc) |>
               (recognize_str_wrap decl))),
          (Some
             ((read_of_poly_variant l y ~decl ~loc) |>
                (read_tag_str_wrap decl))))
    | _ -> (None, None) in
  ((write_decl_of_type decl y), (read_decl_of_type decl y), (json_str decl),
    recognize, read_tag)
let write_case (i, i', l) { Parsetree.pcd_name = pcd_name; pcd_args;_} =
  let (i, i', lhs, rhs) =
    match pcd_args with
    | Pcstr_tuple []|Pcstr_record [] ->
        ((i + 1), i', None,
          {
            Parsetree.pexp_desc =
              (Parsetree.Pexp_apply
                 ((rt "Json_int.write"),
                   [(Asttypes.Nolabel,
                      {
                        Parsetree.pexp_desc =
                          (Parsetree.Pexp_ident
                             {
                               Asttypes.txt = (Longident.Lident "buf");
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             });
                        Parsetree.pexp_loc =
                          (Pervasives.(!) Ast_helper.default_loc);
                        Parsetree.pexp_attributes = []
                      });
                   (Asttypes.Nolabel, (Ast_convenience.int i))]));
            Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
            Parsetree.pexp_attributes = []
          })
    | Pcstr_tuple (_::[] as args) ->
        let v = Ppx_deriving.fresh_var [] in
        (i, (i' + 1), (Some (Ast_convenience.pvar v)),
          (write_tuple_contents [v] args ~tag:i' ~poly:true))
    | Pcstr_tuple args ->
        let vars = fresh_vars (List.length args) in
        (i, (i' + 1), (Some (var_ptuple vars)),
          (write_tuple_contents vars args ~tag:i' ~poly:true))
    | Pcstr_record args ->
        (i, (i' + 1), (Some (pattern_of_record args)),
          (write_body_of_record args ~tag:i')) in
  (i, i',
    ((let open Ast_helper in
        Exp.case (Pat.construct (label_of_constructor pcd_name) lhs) rhs) ::
    l))
let write_decl_of_variant d l =
  ((let (_, _, l) = List.fold_left write_case (0, 0, []) l in
    Ast_helper.Exp.function_ l) |> buf_expand) |> (write_str_wrap d)
let read_case ?decl  (i, i', l) { Parsetree.pcd_name = pcd_name; pcd_args;_}
  =
  let f l =
    Ast_helper.Exp.construct (label_of_constructor pcd_name)
      (match l with
       | [] -> None
       | e::[] -> Some e
       | l -> Some (Ast_helper.Exp.tuple l)) in
  match pcd_args with
  | Pcstr_tuple []|Pcstr_record [] ->
      ((i + 1), i',
        ((Ast_helper.Exp.case
            {
              Parsetree.ppat_desc =
                (Parsetree.Ppat_variant
                   ("Cst", (Some (Ast_convenience.pint i))));
              Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
              Parsetree.ppat_attributes = []
            } (Ast_helper.Exp.construct (label_of_constructor pcd_name) None))
        :: l))
  | Pcstr_tuple pcd_args ->
      let expr = read_tuple_contents ?decl pcd_args ~f in
      let case =
        Ast_helper.Exp.case
          {
            Parsetree.ppat_desc =
              (Parsetree.Ppat_variant
                 ("NCst", (Some (Ast_convenience.pint i'))));
            Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
            Parsetree.ppat_attributes = []
          } expr in
      (i, (i' + 1), (case :: l))
  | Pcstr_record pcd_args ->
      let patt =
        {
          Parsetree.ppat_desc =
            (Parsetree.Ppat_variant
               ("NCst", (Some (Ast_convenience.pint i'))));
          Parsetree.ppat_loc = (Pervasives.(!) Ast_helper.default_loc);
          Parsetree.ppat_attributes = []
        }
      and expr =
        let return e =
          Ast_helper.Exp.construct (label_of_constructor pcd_name) (Some e) in
        read_of_record_raw ?decl pcd_args ~return in
      (i, (i' + 1), ((Ast_helper.Exp.case patt expr) :: l))
let read_decl_of_variant decl l =
  ((let (_, _, l) = List.fold_left (read_case ~decl) (0, 0, []) l
    and e =
      {
        Parsetree.pexp_desc =
          (Parsetree.Pexp_apply
             ((lexer "read_case"),
               [(Asttypes.Nolabel,
                  {
                    Parsetree.pexp_desc =
                      (Parsetree.Pexp_ident
                         {
                           Asttypes.txt = (Longident.Lident "buf");
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         });
                    Parsetree.pexp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.pexp_attributes = []
                  })]));
        Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
        Parsetree.pexp_attributes = []
      } in
    Ast_helper.Exp.match_ e (l @ [tag_error_case ()])) |> buf_expand) |>
    (read_str_wrap decl)
let json_decls_of_variant d l =
  ((write_decl_of_variant d l), (read_decl_of_variant d l), (json_str d),
    None, None)
let write_decl_of_record d l = (write_of_record d l) |> (write_str_wrap d)
let read_decl_of_record d l = (read_of_record d l) |> (read_str_wrap d)
let json_decls_of_record d l =
  check_record_fields l;
  ((write_decl_of_record d l), (read_decl_of_record d l), (json_str d), None,
    None)
let json_str_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (Ast_helper.with_default_loc ptype_loc) @@
    (fun () ->
       match d with
       | { Parsetree.ptype_kind = Ptype_variant l;_} ->
           json_decls_of_variant d l
       | { ptype_kind = Ptype_record l;_} -> json_decls_of_record d l
       | { ptype_manifest = Some y;_} -> json_decls_of_type d y
       | _ ->
           Location.raise_errorf "%s cannot be derived for %s" deriver
             (Ppx_deriving.mangle_type_decl (`Suffix "") d))
let read_sig_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (let s =
     let s = Ppx_deriving.mangle_type_decl (`Suffix "of_json") d in
     Location.mkloc s ptype_loc
   and y =
     let f y =
       {
         Parsetree.ptyp_desc =
           (Parsetree.Ptyp_arrow (Asttypes.Nolabel, lexbuf_t, y));
         Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
         Parsetree.ptyp_attributes = []
       } in
     let y = f (Ppx_deriving.core_type_of_type_decl d) in
     Ppx_deriving.poly_arrow_of_type_decl f d y in
   Ast_helper.Val.mk s y) |> Ast_helper.Sig.value
let recognize_sig_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (let s =
     let s = Ppx_deriving.mangle_type_decl (`Suffix "recognize") d in
     Location.mkloc s ptype_loc
   and y =
     {
       Parsetree.ptyp_desc =
         (Parsetree.Ptyp_arrow
            (Asttypes.Nolabel,
              {
                Parsetree.ptyp_desc =
                  (Parsetree.Ptyp_variant
                     ([Parsetree.Rtag
                         ({
                            Asttypes.txt = "NCst";
                            Asttypes.loc =
                              (Pervasives.(!) Ast_helper.default_loc)
                          }, [], false,
                           [{
                              Parsetree.ptyp_desc =
                                (Parsetree.Ptyp_constr
                                   ({
                                      Asttypes.txt = (Longident.Lident "int");
                                      Asttypes.loc =
                                        (Pervasives.(!)
                                           Ast_helper.default_loc)
                                    }, []));
                              Parsetree.ptyp_loc =
                                (Pervasives.(!) Ast_helper.default_loc);
                              Parsetree.ptyp_attributes = []
                            }]);
                      Parsetree.Rtag
                        ({
                           Asttypes.txt = "Cst";
                           Asttypes.loc =
                             (Pervasives.(!) Ast_helper.default_loc)
                         }, [], false,
                          [{
                             Parsetree.ptyp_desc =
                               (Parsetree.Ptyp_constr
                                  ({
                                     Asttypes.txt = (Longident.Lident "int");
                                     Asttypes.loc =
                                       (Pervasives.(!) Ast_helper.default_loc)
                                   }, []));
                             Parsetree.ptyp_loc =
                               (Pervasives.(!) Ast_helper.default_loc);
                             Parsetree.ptyp_attributes = []
                           }])], Asttypes.Closed, None));
                Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
                Parsetree.ptyp_attributes = []
              },
              {
                Parsetree.ptyp_desc =
                  (Parsetree.Ptyp_constr
                     ({
                        Asttypes.txt = (Longident.Lident "bool");
                        Asttypes.loc =
                          (Pervasives.(!) Ast_helper.default_loc)
                      }, []));
                Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
                Parsetree.ptyp_attributes = []
              }));
       Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
       Parsetree.ptyp_attributes = []
     } in
   Ast_helper.Val.mk s y) |> Ast_helper.Sig.value
let read_with_tag_sig_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (let s =
     let s = Ppx_deriving.mangle_type_decl (`Suffix "of_json_with_tag") d in
     Location.mkloc s ptype_loc
   and y =
     let f y =
       {
         Parsetree.ptyp_desc =
           (Parsetree.Ptyp_arrow (Asttypes.Nolabel, lexbuf_t, y));
         Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
         Parsetree.ptyp_attributes = []
       } in
     let y =
       let y = Ppx_deriving.core_type_of_type_decl d in
       f
         {
           Parsetree.ptyp_desc =
             (Parsetree.Ptyp_arrow
                (Asttypes.Nolabel,
                  {
                    Parsetree.ptyp_desc =
                      (Parsetree.Ptyp_variant
                         ([Parsetree.Rtag
                             ({
                                Asttypes.txt = "NCst";
                                Asttypes.loc =
                                  (Pervasives.(!) Ast_helper.default_loc)
                              }, [], false,
                               [{
                                  Parsetree.ptyp_desc =
                                    (Parsetree.Ptyp_constr
                                       ({
                                          Asttypes.txt =
                                            (Longident.Lident "int");
                                          Asttypes.loc =
                                            (Pervasives.(!)
                                               Ast_helper.default_loc)
                                        }, []));
                                  Parsetree.ptyp_loc =
                                    (Pervasives.(!) Ast_helper.default_loc);
                                  Parsetree.ptyp_attributes = []
                                }]);
                          Parsetree.Rtag
                            ({
                               Asttypes.txt = "Cst";
                               Asttypes.loc =
                                 (Pervasives.(!) Ast_helper.default_loc)
                             }, [], false,
                              [{
                                 Parsetree.ptyp_desc =
                                   (Parsetree.Ptyp_constr
                                      ({
                                         Asttypes.txt =
                                           (Longident.Lident "int");
                                         Asttypes.loc =
                                           (Pervasives.(!)
                                              Ast_helper.default_loc)
                                       }, []));
                                 Parsetree.ptyp_loc =
                                   (Pervasives.(!) Ast_helper.default_loc);
                                 Parsetree.ptyp_attributes = []
                               }])], Asttypes.Closed, None));
                    Parsetree.ptyp_loc =
                      (Pervasives.(!) Ast_helper.default_loc);
                    Parsetree.ptyp_attributes = []
                  }, y));
           Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
           Parsetree.ptyp_attributes = []
         } in
     Ppx_deriving.poly_arrow_of_type_decl f d y in
   Ast_helper.Val.mk s y) |> Ast_helper.Sig.value
let write_sig_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (let s =
     let s = Ppx_deriving.mangle_type_decl (`Suffix "to_json") d in
     Location.mkloc s ptype_loc
   and y =
     let f y =
       {
         Parsetree.ptyp_desc =
           (Parsetree.Ptyp_arrow
              (Asttypes.Nolabel,
                {
                  Parsetree.ptyp_desc =
                    (Parsetree.Ptyp_constr
                       ({
                          Asttypes.txt =
                            (Longident.Ldot
                               ((Longident.Lident "Buffer"), "t"));
                          Asttypes.loc =
                            (Pervasives.(!) Ast_helper.default_loc)
                        }, []));
                  Parsetree.ptyp_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.ptyp_attributes = []
                },
                {
                  Parsetree.ptyp_desc =
                    (Parsetree.Ptyp_arrow
                       (Asttypes.Nolabel, y,
                         {
                           Parsetree.ptyp_desc =
                             (Parsetree.Ptyp_constr
                                ({
                                   Asttypes.txt = (Longident.Lident "unit");
                                   Asttypes.loc =
                                     (Pervasives.(!) Ast_helper.default_loc)
                                 }, []));
                           Parsetree.ptyp_loc =
                             (Pervasives.(!) Ast_helper.default_loc);
                           Parsetree.ptyp_attributes = []
                         }));
                  Parsetree.ptyp_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.ptyp_attributes = []
                }));
         Parsetree.ptyp_loc = (Pervasives.(!) Ast_helper.default_loc);
         Parsetree.ptyp_attributes = []
       } in
     let y = f (Ppx_deriving.core_type_of_type_decl d) in
     Ppx_deriving.poly_arrow_of_type_decl f d y in
   Ast_helper.Val.mk s y) |> Ast_helper.Sig.value
let json_sig_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (let s =
     let s = Ppx_deriving.mangle_type_decl (`Suffix "json") d in
     Location.mkloc s ptype_loc
   and y =
     let f y = rt_t y in
     let y = f (Ppx_deriving.core_type_of_type_decl d) in
     Ppx_deriving.poly_arrow_of_type_decl f d y in
   Ast_helper.Val.mk s y) |> Ast_helper.Sig.value
let sigs_of_decl ({ Parsetree.ptype_loc = ptype_loc;_} as d) =
  (Ast_helper.with_default_loc ptype_loc) @@
    (fun () ->
       let l = [read_sig_of_decl d; write_sig_of_decl d; json_sig_of_decl d] in
       match d with
       | {
           Parsetree.ptype_manifest = Some
             { Parsetree.ptyp_desc = Parsetree.Ptyp_variant _;_};_}
           -> (read_with_tag_sig_of_decl d) :: (recognize_sig_of_decl d) :: l
       | _ -> l)
let register_for_expr s f =
  let core_type ({ Parsetree.ptyp_loc = ptyp_loc;_} as y) =
    let f () = (f y) |> sanitize in Ast_helper.with_default_loc ptyp_loc f in
  let open Ppx_deriving in (create s ~core_type ()) |> register
let _ =
  (register_for_expr "of_json") @@
    (fun y ->
       {
         Parsetree.pexp_desc =
           (Parsetree.Pexp_fun
              (Asttypes.Nolabel, None,
                {
                  Parsetree.ppat_desc =
                    (Parsetree.Ppat_var
                       {
                         Asttypes.txt = "s";
                         Asttypes.loc =
                           (Pervasives.(!) Ast_helper.default_loc)
                       });
                  Parsetree.ppat_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.ppat_attributes = []
                },
                {
                  Parsetree.pexp_desc =
                    (Parsetree.Pexp_apply
                       ((read_of_type y),
                         [(Asttypes.Nolabel,
                            {
                              Parsetree.pexp_desc =
                                (Parsetree.Pexp_apply
                                   ((lexer "init_lexer"),
                                     [(Asttypes.Nolabel,
                                        {
                                          Parsetree.pexp_desc =
                                            (Parsetree.Pexp_apply
                                               ({
                                                  Parsetree.pexp_desc =
                                                    (Parsetree.Pexp_ident
                                                       {
                                                         Asttypes.txt =
                                                           (Longident.Ldot
                                                              ((Longident.Lident
                                                                  "Lexing"),
                                                                "from_string"));
                                                         Asttypes.loc =
                                                           (Pervasives.(!)
                                                              Ast_helper.default_loc)
                                                       });
                                                  Parsetree.pexp_loc =
                                                    (Pervasives.(!)
                                                       Ast_helper.default_loc);
                                                  Parsetree.pexp_attributes =
                                                    []
                                                },
                                                 [(Asttypes.Nolabel,
                                                    {
                                                      Parsetree.pexp_desc =
                                                        (Parsetree.Pexp_ident
                                                           {
                                                             Asttypes.txt =
                                                               (Longident.Lident
                                                                  "s");
                                                             Asttypes.loc =
                                                               (Pervasives.(!)
                                                                  Ast_helper.default_loc)
                                                           });
                                                      Parsetree.pexp_loc =
                                                        (Pervasives.(!)
                                                           Ast_helper.default_loc);
                                                      Parsetree.pexp_attributes
                                                        = []
                                                    })]));
                                          Parsetree.pexp_loc =
                                            (Pervasives.(!)
                                               Ast_helper.default_loc);
                                          Parsetree.pexp_attributes = []
                                        })]));
                              Parsetree.pexp_loc =
                                (Pervasives.(!) Ast_helper.default_loc);
                              Parsetree.pexp_attributes = []
                            })]));
                  Parsetree.pexp_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.pexp_attributes = []
                }));
         Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
         Parsetree.pexp_attributes = []
       })
let _ =
  (register_for_expr "to_json") @@
    (fun y ->
       {
         Parsetree.pexp_desc =
           (Parsetree.Pexp_fun
              (Asttypes.Nolabel, None,
                {
                  Parsetree.ppat_desc =
                    (Parsetree.Ppat_var
                       {
                         Asttypes.txt = "x";
                         Asttypes.loc =
                           (Pervasives.(!) Ast_helper.default_loc)
                       });
                  Parsetree.ppat_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.ppat_attributes = []
                },
                {
                  Parsetree.pexp_desc =
                    (Parsetree.Pexp_let
                       (Asttypes.Nonrecursive,
                         [{
                            Parsetree.pvb_pat =
                              {
                                Parsetree.ppat_desc =
                                  (Parsetree.Ppat_var
                                     {
                                       Asttypes.txt = "buf";
                                       Asttypes.loc =
                                         (Pervasives.(!)
                                            Ast_helper.default_loc)
                                     });
                                Parsetree.ppat_loc =
                                  (Pervasives.(!) Ast_helper.default_loc);
                                Parsetree.ppat_attributes = []
                              };
                            Parsetree.pvb_expr =
                              {
                                Parsetree.pexp_desc =
                                  (Parsetree.Pexp_apply
                                     ({
                                        Parsetree.pexp_desc =
                                          (Parsetree.Pexp_ident
                                             {
                                               Asttypes.txt =
                                                 (Longident.Ldot
                                                    ((Longident.Lident
                                                        "Buffer"), "create"));
                                               Asttypes.loc =
                                                 (Pervasives.(!)
                                                    Ast_helper.default_loc)
                                             });
                                        Parsetree.pexp_loc =
                                          (Pervasives.(!)
                                             Ast_helper.default_loc);
                                        Parsetree.pexp_attributes = []
                                      },
                                       [(Asttypes.Nolabel,
                                          {
                                            Parsetree.pexp_desc =
                                              (Parsetree.Pexp_constant
                                                 (Parsetree.Pconst_integer
                                                    ("50", None)));
                                            Parsetree.pexp_loc =
                                              (Pervasives.(!)
                                                 Ast_helper.default_loc);
                                            Parsetree.pexp_attributes = []
                                          })]));
                                Parsetree.pexp_loc =
                                  (Pervasives.(!) Ast_helper.default_loc);
                                Parsetree.pexp_attributes = []
                              };
                            Parsetree.pvb_attributes = [];
                            Parsetree.pvb_loc =
                              (Pervasives.(!) Ast_helper.default_loc)
                          }],
                         {
                           Parsetree.pexp_desc =
                             (Parsetree.Pexp_sequence
                                ({
                                   Parsetree.pexp_desc =
                                     (Parsetree.Pexp_apply
                                        ((write_of_type y ~poly:false),
                                          [(Asttypes.Nolabel,
                                             {
                                               Parsetree.pexp_desc =
                                                 (Parsetree.Pexp_ident
                                                    {
                                                      Asttypes.txt =
                                                        (Longident.Lident
                                                           "buf");
                                                      Asttypes.loc =
                                                        (Pervasives.(!)
                                                           Ast_helper.default_loc)
                                                    });
                                               Parsetree.pexp_loc =
                                                 (Pervasives.(!)
                                                    Ast_helper.default_loc);
                                               Parsetree.pexp_attributes = []
                                             });
                                          (Asttypes.Nolabel,
                                            {
                                              Parsetree.pexp_desc =
                                                (Parsetree.Pexp_ident
                                                   {
                                                     Asttypes.txt =
                                                       (Longident.Lident "x");
                                                     Asttypes.loc =
                                                       (Pervasives.(!)
                                                          Ast_helper.default_loc)
                                                   });
                                              Parsetree.pexp_loc =
                                                (Pervasives.(!)
                                                   Ast_helper.default_loc);
                                              Parsetree.pexp_attributes = []
                                            })]));
                                   Parsetree.pexp_loc =
                                     (Pervasives.(!) Ast_helper.default_loc);
                                   Parsetree.pexp_attributes = []
                                 },
                                  {
                                    Parsetree.pexp_desc =
                                      (Parsetree.Pexp_apply
                                         ({
                                            Parsetree.pexp_desc =
                                              (Parsetree.Pexp_ident
                                                 {
                                                   Asttypes.txt =
                                                     (Longident.Ldot
                                                        ((Longident.Lident
                                                            "Buffer"),
                                                          "contents"));
                                                   Asttypes.loc =
                                                     (Pervasives.(!)
                                                        Ast_helper.default_loc)
                                                 });
                                            Parsetree.pexp_loc =
                                              (Pervasives.(!)
                                                 Ast_helper.default_loc);
                                            Parsetree.pexp_attributes = []
                                          },
                                           [(Asttypes.Nolabel,
                                              {
                                                Parsetree.pexp_desc =
                                                  (Parsetree.Pexp_ident
                                                     {
                                                       Asttypes.txt =
                                                         (Longident.Lident
                                                            "buf");
                                                       Asttypes.loc =
                                                         (Pervasives.(!)
                                                            Ast_helper.default_loc)
                                                     });
                                                Parsetree.pexp_loc =
                                                  (Pervasives.(!)
                                                     Ast_helper.default_loc);
                                                Parsetree.pexp_attributes =
                                                  []
                                              })]));
                                    Parsetree.pexp_loc =
                                      (Pervasives.(!) Ast_helper.default_loc);
                                    Parsetree.pexp_attributes = []
                                  }));
                           Parsetree.pexp_loc =
                             (Pervasives.(!) Ast_helper.default_loc);
                           Parsetree.pexp_attributes = []
                         }));
                  Parsetree.pexp_loc =
                    (Pervasives.(!) Ast_helper.default_loc);
                  Parsetree.pexp_attributes = []
                }));
         Parsetree.pexp_loc = (Pervasives.(!) Ast_helper.default_loc);
         Parsetree.pexp_attributes = []
       })
let _ =
  let core_type ({ Parsetree.ptyp_loc = ptyp_loc;_} as y) =
    let f () = (json_of_type y) |> sanitize in
    Ast_helper.with_default_loc ptyp_loc f
  and type_decl_str ~options:_  ~path:_  l =
    let (lw, lr, lj, lp, lrv) =
      let f d (lw, lr, lj, lp, lrv) =
        let (w, r, j, p, rv) = json_str_of_decl d in
        ((w :: lw), (r :: lr), (j :: lj),
          (match p with | Some p -> p :: lp | None -> lp),
          (match rv with | Some rv -> rv :: lrv | None -> lrv))
      and acc = ([], [], [], [], []) in List.fold_right f l acc
    and f = Ast_helper.Str.value Asttypes.Recursive
    and f' = Ast_helper.Str.value Asttypes.Nonrecursive in
    let l = [f (lrv @ lr); f lw; f' lj] in
    match lp with | [] -> l | _ -> (f lp) :: l
  and type_decl_sig ~options:_  ~path:_  l =
    (List.map sigs_of_decl l) |> List.flatten in
  let open Ppx_deriving in
    (create "json" ~core_type ~type_decl_str ~type_decl_sig ()) |> register
